# pylint: disable=dangerous-default-value
from typing import Any


class NOT_SET:
    def __bool__(self) -> bool:
        return False

    def __copy__(self) -> Any:
        return NOT_SET

    def __deepcopy__(self, memodict: dict = {}) -> Any:
        return NOT_SET


NOT_SET: Any = NOT_SET()  # type:ignore



# pylint:disable=comparison-with-callable
import json
from typing import Any, Callable, Optional, Tuple, Union, get_origin

from docstring_parser import parse as doc_parse
from flask import jsonify, request
from pydantic import BaseModel, ValidationError, parse_obj_as, schema_of
from werkzeug.routing import parse_rule

from .constants import NOT_SET
from .param import Param, ParamType
from .security import HttpAuthBase


class ApiConfigError(Exception):
    """There is a mistake in the API configuration."""


class Callback(BaseModel):
    name: str
    url: str
    method: str
    request_body: Optional[Any]
    params: Optional[list[Param]]
    response_codes: dict[int, str]

    class Config:
        arbitrary_types_allowed = True


class SerializationModel(BaseModel):
    data: Any

    class Config:
        arbitrary_types_allowed = True


class Operation:
    """Operation represents handler for one endpoint."""

    def __init__(
        self,
        path: str,
        method: str,
        view_func: Callable,
        responses: Optional[Any] = None,
        callbacks: Optional[list[Callback]] = None,
        summary: str = "",
        description: str = "",
        params: Optional[dict[str, Param]] = None,
        auth: Any = NOT_SET,
    ):
        self.path = path
        self.method = method
        self.view_func = view_func
        self.definitions: dict = {}
        self.responses = self._sanitize_responses(responses, self.view_func)
        self.callbacks = callbacks
        self.summary = summary
        self.description = description
        self.auth = auth
        self.params = params or self._parse_params(path)

    def run(self, *args: Any, **kwargs: Any) -> Any:
        # Run authentication if configured
        if self.auth and self.auth() is None:
            return jsonify("Unauthorized"), 401

        try:
            for param_name, param in self.params.items():
                # Parse query params
                if param.param_type == ParamType.QUERY and param_name in request.args:
                    kwargs[param_name] = parse_obj_as(
                        param.model, request.args[param_name]
                    )
                # Parse request body
                elif param.param_type == ParamType.BODY:
                    kwargs[param_name] = parse_obj_as(param.model, request.json)
        except ValidationError as validation_error:
            return validation_error.json(), 400

        # Run the original view function
        resp = self.view_func(*args, **kwargs)

        # Find model and response code for the returned object
        for code, model in self.responses.items():
            # if the mode is some generic with specified type
            # e.g. list[str], dict[str, Response], etc, we can't use isinstance and
            # at first need to get the unspecified generic type - e.g. list, dict, etc
            # TODO match also the inner types of generics - but that's a corner case
            if isinstance(resp, get_origin(model) or model):
                return jsonify(self.serialize(resp)), code

        raise ApiConfigError(f"No response schema matches returned type {type(resp)}")

    @staticmethod
    def _sanitize_responses(responses: Any, view_func: Callable) -> dict[int, Any]:
        func_return_type = view_func.__annotations__.get("return")

        # Return code not specified, setting it to 200
        if not isinstance(responses, dict):
            responses = {200: responses} if responses else {}
        else:
            # convert all response codes to ints
            responses = {int(k): v for k, v in responses.items()}

        # If responses weren't specified, try to generate it from return type
        if not responses:
            # It can't be an Union
            if func_return_type is None or get_origin(func_return_type) == Union:
                raise ApiConfigError("Return type not specified.")
            responses[200] = func_return_type

        # Check if for each returned type there is implicitly or explicitly defined response model and code
        if func_return_type:
            if get_origin(func_return_type) == Union:
                for ret_type in func_return_type.__args__:
                    if ret_type not in responses.values():
                        raise ApiConfigError(
                            f"Return type {ret_type} http code must be specified explicitly."
                        )

            # If we specified different return type as we specified as response
            elif 200 in responses and responses[200] != func_return_type:
                raise ApiConfigError(
                    f"Return type of the function {type(func_return_type)} does not match response type {type(responses[200])}"
                )
        return responses

    @classmethod
    def serialize(cls, resp: Any) -> Any:
        """Convert response object into json serializable object.

        TODO: Avoid json serialization and deserialization.
        """
        return json.loads(SerializationModel(data=resp).json())["data"]

    def _obj_schema(self, obj: Any) -> dict:
        """Generate schema for a model using pydantic and store definitions."""
        schema = schema_of(obj, ref_template="#/components/schemas/{model}")
        self.definitions.update(schema.get("definitions", {}))

        if "definitions" in schema:
            del schema["definitions"]

        return schema

    def get_callback_schema(self, cb: Callback) -> dict:
        schema: dict = {}
        if cb.request_body:
            schema["requestBody"] = {
                "required": True,
                "content": {
                    "application/json": {"schema": self._obj_schema(cb.request_body)}
                },
            }
        if cb.params:
            schema["parameters"] = [param.schema for param in cb.params]

        if cb.response_codes:
            schema["responses"] = {
                str(code): {"description": description}
                for code, description in cb.response_codes.items()
            }

        return {cb.url: {cb.method.lower(): schema}}

    def get_schema(self) -> dict:
        """Generate schema for the operation."""
        doc = doc_parse(self.view_func.__doc__ or "")

        operation_schema: dict[str, Any] = {
            "summary": doc.short_description or self.summary,
            "description": doc.long_description or self.description,
            "responses": {},
        }

        for code, response in self.responses.items():
            schema = self._obj_schema(response)
            operation_schema["responses"][str(code)] = {
                "content": {"application/json": {"schema": schema}},
                "description": "",
            }

        parameters = []

        for param in self.params.values():
            if param.param_type == ParamType.BODY:
                operation_schema["requestBody"] = {
                    "content": {"application/json": {"schema": param.schema}}
                }
            else:
                parameters.append(param.schema)

        if self.auth:
            operation_schema["security"] = [{self.auth.schema_name: []}]

        if parameters:
            operation_schema["parameters"] = parameters

        if self.callbacks:
            callback_schema = {}
            for callback in self.callbacks:
                callback_schema[callback.name] = self.get_callback_schema(callback)
            operation_schema["callbacks"] = callback_schema

        return operation_schema

    def _parse_path_params(
        self, path: str, param_docs: dict[str, str]
    ) -> dict[str, Param]:
        """Parse path params from endpoint path."""
        return {
            variable: Param.from_path(converter, arguments, variable, param_docs)
            for converter, arguments, variable in parse_rule(path)
            if converter is not None
        }

    def _parse_func_params(
        self, param_docs: dict[str, str]
    ) -> Tuple[dict[str, Param], dict[str, Param]]:
        """Parse query and body params from function arguments."""
        body_params = {}
        query_params = {}
        for param_name, param in self.view_func.__annotations__.items():
            if param_name == "return":
                continue
            if get_origin(param) in (list, dict, tuple) or issubclass(param, BaseModel):
                body_params[param_name] = Param(
                    name=param_name,
                    model=param,
                    param_type=ParamType.BODY,
                    schema=self._obj_schema(param),
                )
            else:
                mapper = {str: "string", int: "integer"}

                query_params[param_name] = Param(
                    name=param_name,
                    model=param,
                    param_type=ParamType.QUERY,
                    schema={
                        "name": param_name,
                        "in": ParamType.QUERY.value,
                        "required": False,
                        "schema": {
                            "type": mapper[param],
                        },
                        "description": param_docs.get(param_name, ""),
                    },
                )

        return query_params, body_params

    def _parse_params(self, path: str) -> dict[str, Param]:
        """Parse path, query and body params from endpoint path and function arguments.

        The idea is:
        - if a param is in endpoint path string - it is a path param
        - if it's not and it's a complex type - list, dict, object - it is body param
        - otherwise it is query param
        """
        param_docs = {
            param.arg_name: param.description or ""
            for param in doc_parse(self.view_func.__doc__ or "").params
        }

        path_params = self._parse_path_params(path, param_docs)
        query_params, body_params = self._parse_func_params(param_docs)

        for param in path_params:
            # We recognized the param as path param, not query param
            if param in query_params:
                del query_params[param]
            elif param in body_params:
                raise ApiConfigError(
                    f"Param of type {type(body_params[param])} can't be path param."
                )
            else:
                raise ApiConfigError(f"Function is missing {param} argument")

        if len(body_params) > 1:
            raise ApiConfigError("Multiple complex objects in function arguments.")

        return path_params | query_params | body_params

    def get_openapi_path(self) -> str:
        """Convert flask endpoint path into openapi path."""
        subs = []

        for converter, _, variable in parse_rule(self.path):
            if converter is None:
                subs.append(variable)
                continue
            subs.append(f"{{{variable}}}")
        return "".join(subs)

    def add_prefix(self, prefix: str) -> None:
        self.path = prefix + self.path

    def update_auth(self, auth: Optional[HttpAuthBase] = None) -> None:
        self.auth = self.auth if self.auth != NOT_SET else auth



import abc
from typing import Any, Optional

from flask import request


class HttpAuthBase(abc.ABC):
    openapi_type: str = "http"
    schema_name: str

    @abc.abstractmethod
    def __call__(self) -> Optional[Any]:
        pass  # pragma: no cover

    @abc.abstractmethod
    def schema(self) -> dict:
        pass


class HttpBearer(HttpAuthBase, abc.ABC):

    schema_name: str = "bearerTokenAuth"
    openapi_scheme: str = "bearer"
    header: str = "Authorization"

    def __call__(self) -> Optional[Any]:

        auth_value = request.headers.get(self.header)
        if not auth_value:
            return None
        parts = auth_value.split(" ")

        if parts[0].lower() != self.openapi_scheme:
            return None
        token = " ".join(parts[1:])
        return self.authenticate(token)

    @abc.abstractmethod
    def authenticate(self, token: str) -> Optional[Any]:
        pass  # pragma: no cover

    def schema(self) -> dict:
        return {
            self.schema_name: {"scheme": self.openapi_scheme, "type": self.openapi_type}
        }



from __future__ import annotations

from typing import Any, Callable, Optional

from flask import Flask

from flask_ninja.constants import NOT_SET
from flask_ninja.operation import Callback, Operation
from flask_ninja.param import Param


class Router:
    def __init__(
        self,
        auth: Any = NOT_SET,
        app: Optional[Flask] = None,
        operations: Optional[list[Operation]] = None,
    ):
        self.app = app
        self.auth = auth
        self.operations: list[Operation] = operations or []

    def get(self, path: str, **kwargs: Any) -> Callable:
        return self.add_route("GET", path, **kwargs)

    def post(self, path: str, **kwargs: Any) -> Callable:
        return self.add_route("POST", path, **kwargs)

    def put(self, path: str, **kwargs: Any) -> Callable:
        return self.add_route("PUT", path, **kwargs)

    def patch(self, path: str, **kwargs: Any) -> Callable:
        return self.add_route("PATCH", path, **kwargs)

    def delete(self, path: str, **kwargs: Any) -> Callable:
        return self.add_route("DELETE", path, **kwargs)

    def add_route(
        self,
        method: str,
        path: str,
        responses: Optional[Any] = None,
        auth: Any = NOT_SET,
        summary: str = "",
        description: str = "",
        params: Optional[dict[str, Param]] = None,
        callbacks: Optional[list[Callback]] = None,
    ) -> Callable:
        def decorator(func: Callable) -> Callable:
            operation = Operation(
                path,
                method,
                func,
                responses=responses,
                auth=auth if auth != NOT_SET else self.auth,
                summary=summary,
                description=description,
                params=params,
                callbacks=callbacks,
            )
            self.add_operation(operation)
            return func

        return decorator

    def add_router(self, router: Router, prefix: str = "") -> None:
        for operation in router.operations:
            operation.add_prefix(prefix)
            operation.update_auth(self.auth)
            self.add_operation(operation)

    def add_operation(self, operation: Operation) -> None:
        self.operations.append(operation)
        if self.app:
            self.app.add_url_rule(
                operation.path,
                operation.view_func.__name__,
                operation.run,
                methods=[operation.method],
            )



from __future__ import annotations

from enum import Enum
from typing import Any, Optional

from werkzeug.routing import parse_converter_args


class ParamType(Enum):
    BODY = "body"
    PATH = "path"
    QUERY = "query"


class Param:
    """Stores one endpoint parameter and its schema."""

    name: str
    model: Any
    param_type: ParamType
    schema: dict

    def __init__(
        self,
        name: str,
        model: Any,
        param_type: ParamType,
        schema: Optional[dict] = None,
        description: str = "",
        required: bool = False,
    ):
        self.name = name
        self.model = model
        self.param_type = param_type
        self.schema = schema or self.default_schema(
            model, param_type, name, description, required
        )

    @classmethod
    def from_path(
        cls, converter: str, arguments: Any, variable: str, param_docs: dict[str, str]
    ) -> Param:
        """Parse flask endpoint param and generate openapi schema for it."""
        args, kwargs = [], {}

        if arguments:
            args, kwargs = parse_converter_args(arguments)

        schema: dict[str, Any] = {}
        if converter == "any":
            schema = {
                "type": "string",
                "enum": args,
            }
        elif converter == "int":
            schema = {
                "type": "integer",
            }
            if "max" in kwargs:
                schema["maximum"] = kwargs["max"]
            if "min" in kwargs:
                schema["minimum"] = kwargs["min"]
        elif converter == "float":
            schema = {
                "type": "number",
                "format": "float",
            }
        elif converter == "uuid":
            schema = {
                "type": "string",
                "format": "uuid",
            }
        elif converter == "path":
            schema = {
                "type": "string",
                "format": "path",
            }
        elif converter == "string":
            schema = {
                "type": "string",
            }
            for prop in ["length", "maxLength", "minLength"]:
                if prop in kwargs:
                    schema[prop] = kwargs[prop]
        elif converter == "default":
            schema = {"type": "string"}

        description = param_docs.get(variable, "")

        return cls(
            name=variable,
            model=int if schema["type"] == "integer" else str,
            param_type=ParamType.PATH,
            schema={
                "name": variable,
                "in": ParamType.PATH.value,
                "required": True,
                "schema": schema,
                "description": description,
            },
        )

    @staticmethod
    def default_schema(
        model: Any, param_type: ParamType, name: str, description: str, required: bool
    ) -> dict:
        return {
            "name": name,
            "in": param_type.value,
            "required": required,
            "schema": {"type": "integer"} if model == int else {"type": "string"},
            "description": description,
        }



from flask_ninja.api import NinjaAPI, Server
from flask_ninja.operation import ApiConfigError, Callback, Operation
from flask_ninja.param import Param, ParamType
from flask_ninja.router import Router
from flask_ninja.security import HttpAuthBase, HttpBearer



import re
from typing import Any, Callable, Optional

from flask import Blueprint, Flask, render_template
from pydantic import BaseModel

from .constants import NOT_SET
from .router import Router
from .swagger_ui import swagger_ui_path


class Server(BaseModel):
    url: str
    description: str


class NinjaAPI:
    OPENAPI_VERSION = "3.0.3"

    def __init__(
        self,
        app: Flask,
        auth: Any = NOT_SET,
        title: str = "",
        description: str = "",
        version: str = "1.0.0",
        servers: Optional[list[Server]] = None,
    ):
        swagger_bp = Blueprint(
            "swagger_ui",
            __name__,
            static_url_path="",
            static_folder=swagger_ui_path,
            template_folder=swagger_ui_path,
        )
        swagger_bp.add_url_rule(
            "/",
            "docs",
            lambda: render_template("index.j2", openapi_spec_url="/openapi.json"),
        )
        app.register_blueprint(swagger_bp, url_prefix="/docs")
        app.add_url_rule("/openapi.json", "openapi", self.get_schema, methods=["GET"])
        self.router = Router(auth=auth, app=app)
        self.title = title
        self.description = description
        self.version = version
        self.servers = servers
        self.model_definitions: dict[str, Any] = {}

    def get(self, path: str, **kwargs: Any) -> Callable:
        return self.router.add_route("GET", path, **kwargs)

    def post(self, path: str, **kwargs: Any) -> Callable:
        return self.router.add_route("POST", path, **kwargs)

    def put(self, path: str, **kwargs: Any) -> Callable:
        return self.router.add_route("PUT", path, **kwargs)

    def patch(self, path: str, **kwargs: Any) -> Callable:
        return self.router.add_route("PATCH", path, **kwargs)

    def delete(self, path: str, **kwargs: Any) -> Callable:
        return self.router.add_route("DELETE", path, **kwargs)

    def add_router(self, router: Router, prefix: str = "") -> None:
        self.router.add_router(router, prefix)

    def get_schema(self) -> dict:
        """Generate Openapi schema for the API."""

        paths: dict = {}
        security_schemes: dict = {}

        for operation in self.router.operations:
            # Arguments in paths has format e.g. <int:id> in flask which is different than in openapi e.g. {id}
            # therefore we need to convert it to openapi format
            swagger_path = operation.get_openapi_path()
            if swagger_path not in paths:
                paths[swagger_path] = {}
            paths[swagger_path][operation.method.lower()] = operation.get_schema()

            # Pydantic somehow caches generated definitions and second time a schema is generated for a model
            # only a reference is returned. Therefore we need to carefully store the generated definitions
            # to be able to provide them in the response.
            self.model_definitions.update(operation.definitions)

            if operation.auth:
                security_schemes.update(operation.auth.schema())

        schema: dict = {
            "openapi": self.OPENAPI_VERSION,
            "info": {
                "title": self.title,
                # strip whitespaces in the beginning of lines caused by indents
                "description": re.sub(r"\n *", "\n", self.description),
                "version": self.version,
            },
            "components": {
                "schemas": self.model_definitions,
                "securitySchemes": security_schemes,
            },
            "paths": paths,
        }
        if self.servers:
            schema["servers"] = [server.dict() for server in self.servers]

        return schema



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['flask_ninja']

package_data = \
{'': ['*'], 'flask_ninja': ['swagger-ui-4.12.0/*']}

install_requires = \
['Flask>=1.1.2', 'docstring-parser>=0.14.1,<0.15.0', 'pydantic>=1.9.0,<2.0.0']

setup_kwargs = {
    'name': 'flask-ninja',
    'version': '1.1.2',
    'description': 'Flask Ninja is a web framework for building APIs with Flask and Python 3.9+ type hints.',
    'long_description': '# Flask Ninja\n\n**Flask Ninja** is a web framework for building APIs with Flask and Python 3.9+ type hints.\n\nKey features:\n\n- Easy: Designed to be easy to use and intuitive.\n- Fast to code: Type hints and automatic docs lets you focus only on business logic.\n- Standards-based: Based on the open standards for APIs: OpenAPI (previously known as Swagger) and JSON Schema.\n- Models based: Pydantic models support and automatic (de)serialization of requests/responses.\n- Secure: Natively supports various authentication methods for the requests.\n\n## Installation\n\n```\npip install flask-ninja\n```\n\n## Usage\n\nIn your flask project where you create flask app:\n\n```Python\nfrom flask import Flask\nfrom flask_ninja import NinjaAPI\nfrom pydantic import BaseModel\n\napp = Flask(__name__)\napi = NinjaAPI(app)\n\nclass Response(BaseModel):\n    """Response model containing results of various number operations."""\n    sum: int\n    difference: int\n    product: int\n    power: int\n\n@api.get("/compute")\ndef compute(a: int, b: int) -> Response:\n    """Computes results of various number operations.\n\n    This endpoint returns a result of the following operations:\n    - sum\n    - difference\n    - product\n    - power\n\n    :param int a: First number\n    :param int b: Second number number\n    """\n    return Response(\n        sum=a + b,\n        difference=a - b,\n        product=a * b,\n        power=a ** b\n    )\n\nif __name__ == "__main__":\n    app.run()\n```\n\n**That\'s it !**\n\nNow you\'ve just created an API that:\n\n- receives an HTTP GET request at `/compute`\n- takes, validates and type-casts GET parameters `a` and `b`\n- validates the returned Response object and serializes it into JSON\n- generates an OpenAPI schema for defined operation\n\n### Interactive API docs\n\nNow go to <a href="http://127.0.0.1:8000/docs" target="_blank">http://127.0.0.1:5000/docs</a>\n\nYou will see the automatic interactive API documentation (provided by <a href="https://github.com/swagger-api/swagger-ui" target="_blank">Swagger UI</a>):\n',
    'author': 'Michal Korbela',
    'author_email': 'michal.korbela@kiwi.com',
    'maintainer': None,
    'maintainer_email': None,
    'url': None,
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.9,<4.0',
}


setup(**setup_kwargs)



import os


def get_path(rel: str) -> str:
    return os.path.join(
        os.path.abspath(os.path.dirname(os.path.realpath(__file__))), rel
    )


swagger_ui_4_12_0_path = get_path("swagger-ui-4.12.0")

swagger_ui_path = swagger_ui_4_12_0_path
